
const PATCH_IDENTIFIER = 0x3561706D;
const PATTERN_IDENTIFIER = 0xF3559321;
const SEQUENCE_IDENTIFIER = 0xFE099312;
const NAME_LENGTH = 16;


const SYSEX_HEADER_BYTES  = 12;
const BANK_OF_PATCHES_LENGTH = 50;
const BANK_OF_PATTERNS_LENGTH = 8;
const BANK_OF_SEQUENCES_LENGTH = 8;

const PATCH_LENGTH_BYTES_UNPACKED = 4096;
const PATTERN_LENGTH_BYTES_UNPACKED = 2048; //192 // using sizeof
const SEQUENCE_LENGTH_BYTES_UNPACKED = (4096 * (5 + 8));

// ensure buffer large enough to hold bytes for any transfer
const LARGEST_BUFFER_LENGTH = (PATCH_LENGTH_BYTES_UNPACKED  * BANK_OF_PATCHES_LENGTH * 2);

const MAX_PACKET_LENGTH = 64;


const 	 MIDI_SYSEX_NOT_SET  						= 0;
const    MIDI_SYSEX_REQUEST_BITMAP 					= 10;
const    MIDI_SYSEX_REQUEST_OSCILLOSCOPE_TRACE 		= 11;
const    MIDI_SYSEX_REQUEST_BULK_FLASH_DATA_SEND 	= 12;
const    MIDI_SYSEX_RECEIVE_BULK_FLASH_DATA 		= 13;

const    MIDI_SYSEX_REQUEST_GLOBALS 		= 15;

const    MIDI_SYSEX_REQUEST_PATCH 			= 20;
const    MIDI_SYSEX_REQUEST_PATCHES_BANK	= 21;

const    MIDI_SYSEX_REQUEST_PATTERN 		= 25;
const    MIDI_SYSEX_REQUEST_PATTERN_BANK 	= 26;

const    MIDI_SYSEX_REQUEST_SEQUENCE 		= 30;
const    MIDI_SYSEX_REQUEST_SEQUENCE_BANK	= 31;


const    MIDI_SYSEX_BITMAP_TRANSFER 		= 40;
const    MIDI_SYSEX_OSCILLOSCOPE_TRANSFER 	= 41;
const    MIDI_SYSEX_GLOBALS_TRANSFER 		= 45;

const    MIDI_SYSEX_PATCH_TRANSFER 			= 50;
const    MIDI_SYSEX_PATCH_BANK_TRANSFER		= 51;
const    MIDI_SYSEX_PATTERN_TRANSFER 		= 55;
const    MIDI_SYSEX_PATTERN_BANK_TRANSFER	= 56;
const    MIDI_SYSEX_SEQUENCE_TRANSFER 		= 60;
const    MIDI_SYSEX_SEQUENCE_BANK_TRANSFER 	= 61;

const    MIDI_SYSEX_FIRMWARE 				= 70; // 0x32
const    MIDI_SYSEX_START_TRANSFER 			= 71;
const    MIDI_SYSEX_CONTINUE_TRANSFER		= 72;
const    MIDI_SYSEX_END_TRANSFER			= 73; //0x35
const    MIDI_SYSEX_START_AND_END_TRANSFER 	= 74; // used for small transfers
const    MIDI_SYSEX_ABORT_TRANSFER			= 75;



const LENGTH_OF_PATCH_NAME = 16;


const  BYTE_SYSEX_START   = 0xF0;
const  BYTE_SYSEX_END     = 0xF7;

const  BYTE_SYSEX_ID1     = 0x00; // Motas Electronics Limited assigned MIDI SYSEX IDs
const  BYTE_SYSEX_ID2     = 0x21; // Motas Electronics Limited assigned MIDI SYSEX IDs
const  BYTE_SYSEX_ID3     = 0x2F; // Motas Electronics Limited assigned MIDI SYSEX IDs


function convertMIDIbytes(currentByte,  i)
{
    var outVal = -1;
    if ((i % 8) == 0) //  this byte is the MSB for next 7 bytes, since cannot transmit values > 0x7F over MIDI
    {
        msbByte = currentByte;
        msbCount = 1;
    } else
    {
        outVal = currentByte | ((msbByte << msbCount) & 0x80);
        msbCount++;
    }
    return outVal;
}


function stringFromArray(data)
{
  var count = data.length;
  var str = "";
  
  for(var index = 0; index < count; index += 1)
    str += String.fromCharCode(data[index]);
  
  return str;
}


function sendMIDINote(id)
{	
	var message = [];
    message[0] = 0x80;
    message[1] = 46;
    message[2] = 56
	var output = MIDIOutDevice.outputs.get(id);
	output.send(message); 
}

function sendMIDISysExCommand(id, model, channel, msg1,  msg2)
{
    //debugMessage('MIDI data', 'send sysEx command');

    var message = [];
    message[0] = BYTE_SYSEX_START;
    message[1] = BYTE_SYSEX_ID1;
    message[2] = BYTE_SYSEX_ID2;

    message[3] = BYTE_SYSEX_ID3;
    message[4] = model;
    message[5] = channel;

    message[6] = 0;
    message[7] = msg1;
    message[8] = msg2;

    message[9] = BYTE_SYSEX_END;

	MIDIOutDevice.outputs.get(id).send(message);
	

	//debugMessage('data', message);
}


function drawBitmap(buffer)
{

	var c = document.getElementById("myCanvas");

	var ctx = c.getContext("2d");
	var pixels = ctx.createImageData(128, 64);
    
	
	var colour = document.getElementById("colourPicker").value;
	
	
	
    var x = 0;
    var y = 0;
    var i = 0;
    for (i = 0; i < 4096; i++)
    {
        for (var n = 0; n < 2; n++)
        {
            var val  = buffer[i];
            var data ;
            if (n == 0)
            {
                data = (val & 0x0F) ;
            }
            else
            {
                data = ((val >> 4) & 0x0F) ;
            }
            data <<= 4;
            
            // compress so dark stuff is not too dark
            data = Math.floor((data + 1) * 3 + 1);
            
            if (data > 0xFF)
            	data = 0xFF;
            else if (data < 0)
            	data = 0;
            

            if (data > 0xFF)
                data = 0xFF;
            var index = (x % 128 + y * 128) * 4;
			pixels.data[index] 	   = (data * parseInt(colour.substr(1, 2), 16)) /256;
			pixels.data[index + 1] = (data * parseInt(colour.substr(3, 2), 16)) /256;
			pixels.data[index + 2] = (data * parseInt(colour.substr(5, 2), 16)) /256;
			pixels.data[index + 3] = 255;
           // line[x  % 128] = (data) | (data << 8) | ((quint32) data << 16); // white

            x++;
            if ((x % 128 ) == 0)
            {
                y++;
            }
        }
    }

	ctx.putImageData(pixels, 0, 0);
	
	//ctx.scale(3, 3);
	ctx.drawImage(c, 0, 0);

	
	var cL = document.getElementById("myCanvasLarge");
	ctx = cL.getContext("2d");
	ctx.drawImage(c, 0, 0);
	
	screenShotUpdated = true;
}



function convertRawMIDI( buf, totalSize)
{
    if (totalSize < 128)
    {
        //qDebug() << "ERROR";
    	debugMessage('ERROR', 'size < 128 long!');
        return 0 ;
    }

    var bufIndex = SYSEX_HEADER_BYTES;
    var convertedIndex = 0;
    var index = 0;
    while (bufIndex < totalSize)
    {
        // decode the MIDI buffer to the original form (with values > 0x7F)
        if (buf[bufIndex] == 0xF7)
        {
            // restart
            index = 0;
            bufIndex += SYSEX_HEADER_BYTES + 1;
        }
        if (bufIndex >= totalSize)
            break;
        var c = convertMIDIbytes(buf[bufIndex], index);
        if (c >= 0)
        {
            buf[convertedIndex++] = c;
        }
        index++;
        bufIndex++;
     }

    return convertedIndex;
}


function read32( offset,  buffer)
{
    var val = 0;
    for (var i = 0; i < 4; i++)
    {
        val += ( buffer[offset + i]) << (8 * (i));
    }
    return val >>> 0; // convert to UINT32
}


function memset(buf, value, length)
{
	var i;
	for (i = 0; i < length; i++)
	{
		buf[i] = value;
	}
}
function write32( val,  offset,  buffer)
{
    for (var i = 0; i < 4; i++)
    {
       buffer[offset + i]  = (val >>> (8 * i)) & 0xFF;
    }
}


function processMIDIdata(type,  buffer)
{
	
	debugMessage("process MIDI data");
	
	    var convertedLength = convertRawMIDI(buffer, buffer.length);

    var expectedLength = 0;
    var fullBankTransfer = false;
	
    switch (type)
    {
        case MIDI_SYSEX_BITMAP_TRANSFER:
            expectedLength = 4096;
            break;
        case MIDI_SYSEX_OSCILLOSCOPE_TRANSFER:
            expectedLength = OSCILLOSCOPE_TRACE_NUM_POINTS*2;
            break;
        case MIDI_SYSEX_REQUEST_BULK_FLASH_DATA_SEND:

          //  sectorId =  buffer[SYSEX_HEADER_BYTES - 3] * 128 + buffer[SYSEX_HEADER_BYTES - 2];
            expectedLength = MASS_FLASH_TRANSFER_LENGTH_BYTES;
            break;
        case MIDI_SYSEX_PATCH_TRANSFER:
            expectedLength = PATCH_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATCH_BANK_TRANSFER:
            fullBankTransfer = true;
            expectedLength  = BANK_OF_PATCHES_LENGTH * PATCH_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATTERN_TRANSFER:
            expectedLength  = PATTERN_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
            fullBankTransfer = true;
            expectedLength  = BANK_OF_PATTERNS_LENGTH * PATTERN_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_SEQUENCE_TRANSFER:
            expectedLength  =  SEQUENCE_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
            fullBankTransfer = true;
            expectedLength  = BANK_OF_SEQUENCES_LENGTH * SEQUENCE_LENGTH_BYTES_UNPACKED;
            break;

        case MIDI_SYSEX_GLOBALS_TRANSFER:
            expectedLength = SETUP_DATA_LENGTH_BYTES;
            break;
        default:
        	debugMessage("unknown incoming type");
            return;
            break;
    }
    expectedLength += 4;// add on the CRC bytes


	//debugMessage('array length', convertedLength);
	//debugMessage('array length expected', expectedLength);
    if (convertedLength != expectedLength)
    {
 		alert("incorrect transfer length");

        return;
    }
   
    // calculate CRC
    var CRC32  = read32(convertedLength - 4, buffer);

    crcInit();
    var crcVal = crcFastSTMF4(buffer, 0, convertedLength - 4); 

    var fileNameDef = "/";

	//debugMessage("crcVals: ", crcVal, CRC32);
	
    if (CRC32 == crcVal)
    {
        var row = 0;
        var end = 1;


        switch (type)
        {
            case MIDI_SYSEX_BITMAP_TRANSFER:
				//debugMessage("bitmap", "received");
				drawBitmap(buffer);
				//functionRequestScreenShot();
                break;
            case MIDI_SYSEX_OSCILLOSCOPE_TRANSFER:
                updateOscilloscope(buffer);
                break;
/*
            case MIDI_SYSEX_REQUEST_BULK_FLASH_DATA_SEND:
            {
                // write to disk file

                if (TRUE)
                {
                    var fileName =  QFileDialog::getSaveFileName(this,"save FLASH memory from Motas to file",  fileNameDef + "bulk.dat");

                    if (fileName.isEmpty())
                        return;

                    QFileInfo fileInfo(fileName);
                    fileNameLocationDefault = fileInfo.absolutePath() ;




                    qDebug() << fileName;
                    QFile f(fileName);
                    f.open(QIODevice::ReadWrite);
                    f.write((char*) buffer, convertedLength - 4); // don't write the CRC at the end

                    f.close();
                }

            }
                break;
            case MIDI_SYSEX_GLOBALS_TRANSFER:
            {
                // write to disk file

                if (TRUE)
                {
                    QString fileName =  QFileDialog::getSaveFileName(this,"save setup data from Motas to file", fileNameDef +  "setupData.setup");

                    if (fileName.isEmpty())
                        return;
                    QFileInfo fileInfo(fileName);
                    fileNameLocationDefault = fileInfo.absolutePath() ;
                    qDebug() << fileName;
                    QFile f(fileName);
                    f.open(QIODevice::ReadWrite);
                    f.write((char*) buffer, convertedLength - 4); // don't write the CRC at the end

                    f.close();
                }

            }
            break;*/
            case MIDI_SYSEX_PATCH_TRANSFER:
            case MIDI_SYSEX_PATCH_BANK_TRANSFER:
            {
                if (fullBankTransfer)
                {
                    row = 0;
                    end = BANK_OF_PATCHES_LENGTH;
                } else
                {
                     row = getActiveRow();
                }
                for (var i = 0; i < end; i++)
                {
                    
                	bankOfPatches[row] = buffer.slice(i * PATCH_LENGTH_BYTES_UNPACKED, (i + 1) * PATCH_LENGTH_BYTES_UNPACKED);

                    updatePatchUI(row, false, i == end - 1, MIDI_SYSEX_PATCH_TRANSFER);
                    row++;
                }
                break;
            }
            
            
            case MIDI_SYSEX_PATTERN_TRANSFER:
            case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
                if (fullBankTransfer)
                {
                    row = 0;
                    end = BANK_OF_PATTERNS_LENGTH;
                } else
                {
                	row = getActiveRow();
                }
                for (var i = 0; i < end; i++)
                {   
                    bankOfPatterns[row] = buffer.slice(i * PATTERN_LENGTH_BYTES_UNPACKED, (i + 1) * PATTERN_LENGTH_BYTES_UNPACKED);
                    updatePatchUI(row, false, i == end - 1, MIDI_SYSEX_PATTERN_TRANSFER);
                    row++;
                }
                break;
            case MIDI_SYSEX_SEQUENCE_TRANSFER:
            case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
                if (fullBankTransfer)
                {
                    row = 0;
                    end = BANK_OF_SEQUENCES_LENGTH;
                } else
                {
                	row = getActiveRow();
                }
                for (var i = 0; i < end; i++)
                {
                	bankOfSequences[row] = buffer.slice(i * SEQUENCE_LENGTH_BYTES_UNPACKED, (i + 1) * SEQUENCE_LENGTH_BYTES_UNPACKED);

                    updatePatchUI(row, false, i == end - 1, MIDI_SYSEX_SEQUENCE_TRANSFER);

                    row++;
                }
                break;
            default:
                break;
        }
    } else
    {
		debugMessage("CRC error", "wrong");
    }
}

/*
 * 
 * Take the data (already converted to be compatible with MIDI protocol) and package up with 
 * SYSEX MIDI bytes
 */
function lowLevelCreateMIDIStream(buf,  size,   type,  startStopContinue,
         msg0,  msg1,  msg2,  outBuf,  model,  channel, bufIndex)
{
    var d = {evnt0:0, value1:0, value2:0};
    var i = -4;

    var remainder = size % 3;
    var wholePackets = Math.trunc(size / 3);
    var packetsSent = 0;
    var outCount = bufIndex;
    while (1)
    {
     //   d.cin_cable = 0x04; // start/continue sysex
        if (i == -4)
        {
            d.evnt0 = 0xF0; // start sysex
            d.value1 = BYTE_SYSEX_ID1;
            d.value2 = BYTE_SYSEX_ID2;
            i++;
        }
        else if (i == -3)
        {
            d.evnt0 =  BYTE_SYSEX_ID3;
            d.value1 = model;
            d.value2 = channel & 0x0F;
            i++;
        }
        else if (i == -2)
        {
            d.evnt0 = 0;
            d.value1 = type;
            d.value2 = startStopContinue;
            i++;
        }
        else if (i == -1 && startStopContinue != MIDI_SYSEX_START_TRANSFER 
        		&& startStopContinue != MIDI_SYSEX_START_AND_END_TRANSFER)
        {
            i++;
            continue; // skip a transfer
        }
        else if (i == -1)
        {
            // other meta-data
            d.evnt0 = msg0;
            d.value1 = msg1;
            d.value2 = msg2;
            i++;
        }
        else if (packetsSent < wholePackets)
        {
            d.evnt0 = buf[i++];
            d.value1 = buf[i++];
            d.value2 = buf[i++];
            packetsSent++;
        } else // final byte
        {
            break;
        }

        outBuf[outCount++] = d.evnt0;
        outBuf[outCount++] = d.value1;
        outBuf[outCount++] = d.value2;
    }

    // final packet
    if (remainder == 0)
    {
    //    d.cin_cable = 0x05; // end sysex
        d.evnt0 = 0xF7;
        d.value1 = 0;
        d.value2 = 0;
        outBuf[outCount++] = d.evnt0;
    }
    else if (remainder == 1)
    {
     //   d.cin_cable = 0x06; // end sysex
        d.evnt0 = buf[i] ;
        d.value1 = 0xF7;
        d.value2 = 0;
        outBuf[outCount++] = d.evnt0;
        outBuf[outCount++] = d.value1;
    } else if (remainder == 2)
    {
      //  d.cin_cable = 0x07; // end sysex
        d.evnt0 = buf[i++] ;
        d.value1 = buf[i] ;
        d.value2 = 0xF7;
        outBuf[outCount++] = d.evnt0;
        outBuf[outCount++] = d.value1;
        outBuf[outCount++] = d.value2;
    }

    return outCount;
}


/*
 * 
 * Convert the binary data into a form suitable for sending over MIDI
 */

function prepareSysExPacketsFromUnpacked(outBuf, buffer,  totalDataSize,  type,
		msg0,  msg1,  msg2,  model,  channel)
{
//	#define MAX_PACKET_LENGTH 64
	var bytesProcessed = 0;
	var outBuffer = new Uint8Array(MAX_PACKET_LENGTH  + ((MAX_PACKET_LENGTH  + 6) / 7));

	var bytesToSend = 0;
	var startStopContinue = MIDI_SYSEX_START_TRANSFER;

	var bufferSize = totalDataSize;

//	calculate CRC32

	if (totalDataSize % 4 != 0)
	{
		window.alert("Error in data - size must be divisible by 4.");
		return 0;
	}

	crcInit();
	var crc32 = crcFastSTMF4(buffer, 0,  totalDataSize); // calculate CRC of all
	
	totalDataSize += 4;// make room for CRC32;

	do
	{
		var k = 0;
		var packetLength = totalDataSize - bytesProcessed;
		if (packetLength > MAX_PACKET_LENGTH)
			packetLength = MAX_PACKET_LENGTH;
		var transferSize = packetLength  + Math.trunc((packetLength  + 6) / 7);

//		convert to data suitable for MIDI;
		while (k < transferSize) // there is an MSB byte to be added for every 7
			// bytes, since cannot transmit values > 0x7F
		{
			var msb = 0;
			var j;
			for (j = 0; j < 7; j++)
			{
				var offset = bytesProcessed + j;
				if (offset >= totalDataSize)
					break;
				if (offset < bufferSize)
				{
					if (buffer[offset] > 0x7F)
						msb |= (0x40 >>> j);
				} else
				{
					if (((crc32 >>> ((offset - bufferSize) * 8)) & 0xFF) > 0x7F)
						msb |= (0x40 >>> j);
				}
			}
			outBuffer[k++] = msb;
			for (j = 0; j < 7; j++)
			{
				if (bytesProcessed >= totalDataSize)
					break;
				if (k >= transferSize)
					break;

				if ((bytesProcessed) < bufferSize)
				{
					outBuffer[k++] =  buffer[bytesProcessed] & 0x7F;
				} else
				{
					outBuffer[k++] =  (crc32 >>> (bytesProcessed - bufferSize) * 8) & 0x7F;
				}
				bytesProcessed++;
			}
		}
		if (bytesProcessed >= totalDataSize)
		{
			startStopContinue = MIDI_SYSEX_END_TRANSFER;
			if (totalDataSize < MAX_PACKET_LENGTH)
			{
				startStopContinue = MIDI_SYSEX_START_AND_END_TRANSFER;
				//qDebug() << "Sending single packet";
			}
		}
		else if (bytesProcessed > packetLength)
			startStopContinue = MIDI_SYSEX_CONTINUE_TRANSFER;
		
		
		
	//	debugMessage("bytesToSend:"  + bytesToSend);
		if (bytesToSend == 0)
		{
	//		debugMessage("output buffer:"  + outBuffer);
		}
		bytesToSend = lowLevelCreateMIDIStream(outBuffer, transferSize, type, startStopContinue, msg0, msg1, msg2,  outBuf,
				model, channel, bytesToSend);
		
	} while (bytesProcessed < totalDataSize);

	return bytesToSend;
}







